/*!
 * vue-form-container v0.0.3
 * (c) 2018-present Vitor Luiz Cavalcanti <vitorluizc@outlook.com> (https://vitorluizc.github.io)
 * Released under the MIT License.
 */
(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
  typeof define === 'function' && define.amd ? define(factory) :
  (global.VueFormContainer = factory());
}(this, (function () { 'use strict';

  /*!
   * isobject <https://github.com/jonschlinkert/isobject>
   *
   * Copyright (c) 2014-2017, Jon Schlinkert.
   * Released under the MIT License.
   */

  var isobject = function isObject(val) {
    return val != null && typeof val === 'object' && Array.isArray(val) === false;
  };

  /*!
   * get-value <https://github.com/jonschlinkert/get-value>
   *
   * Copyright (c) 2014-2018, Jon Schlinkert.
   * Released under the MIT License.
   */



  var getValue = function(target, path, options) {
    if (!isobject(options)) {
      options = { default: options };
    }

    if (!isValidObject(target)) {
      return typeof options.default !== 'undefined' ? options.default : target;
    }

    if (typeof path === 'number') {
      path = String(path);
    }

    const isArray = Array.isArray(path);
    const isString = typeof path === 'string';
    const splitChar = options.separator || '.';
    const joinChar = options.joinChar || (typeof splitChar === 'string' ? splitChar : '.');

    if (!isString && !isArray) {
      return target;
    }

    if (isString && path in target) {
      return isValid(path, target, options) ? target[path] : options.default;
    }

    let segs = isArray ? path : split(path, splitChar, options);
    let len = segs.length;
    let idx = 0;

    do {
      let prop = segs[idx];
      if (typeof prop === 'number') {
        prop = String(prop);
      }

      while (prop && prop.slice(-1) === '\\') {
        prop = join([prop.slice(0, -1), segs[++idx] || ''], joinChar, options);
      }

      if (prop in target) {
        if (!isValid(prop, target, options)) {
          return options.default;
        }

        target = target[prop];
      } else {
        let hasProp = false;
        let n = idx + 1;

        while (n < len) {
          prop = join([prop, segs[n++]], joinChar, options);

          if ((hasProp = prop in target)) {
            if (!isValid(prop, target, options)) {
              return options.default;
            }

            target = target[prop];
            idx = n - 1;
            break;
          }
        }

        if (!hasProp) {
          return options.default;
        }
      }
    } while (++idx < len && isValidObject(target));

    if (idx === len) {
      return target;
    }

    return options.default;
  };

  function join(segs, joinChar, options) {
    if (typeof options.join === 'function') {
      return options.join(segs);
    }
    return segs[0] + joinChar + segs[1];
  }

  function split(path, splitChar, options) {
    if (typeof options.split === 'function') {
      return options.split(path);
    }
    return path.split(splitChar);
  }

  function isValid(key, target, options) {
    if (typeof options.isValid === 'function') {
      return options.isValid(key, target);
    }
    return true;
  }

  function isValidObject(val) {
    return isobject(val) || Array.isArray(val) || typeof val === 'function';
  }

  function isObjectObject(o) {
    return isobject(o) === true
      && Object.prototype.toString.call(o) === '[object Object]';
  }

  var isPlainObject = function isPlainObject(o) {
    var ctor,prot;

    if (isObjectObject(o) === false) return false;

    // If has modified constructor
    ctor = o.constructor;
    if (typeof ctor !== 'function') return false;

    // If has modified prototype
    prot = ctor.prototype;
    if (isObjectObject(prot) === false) return false;

    // If constructor does not have an Object-specific method
    if (prot.hasOwnProperty('isPrototypeOf') === false) {
      return false;
    }

    // Most likely a plain Object
    return true;
  };

  function set(target, path, value, options) {
    if (!isObject(target)) {
      return target;
    }

    let opts = options || {};
    const isArray = Array.isArray(path);
    if (!isArray && typeof path !== 'string') {
      return target;
    }

    let merge = opts.merge;
    if (merge && typeof merge !== 'function') {
      merge = Object.assign;
    }

    const keys = isArray ? path : split$1(path, opts);
    const len = keys.length;
    const orig = target;

    if (!options && keys.length === 1) {
      result(target, keys[0], value, merge);
      return target;
    }

    for (let i = 0; i < len; i++) {
      let prop = keys[i];

      if (!isObject(target[prop])) {
        target[prop] = {};
      }

      if (i === len - 1) {
        result(target, prop, value, merge);
        break;
      }

      target = target[prop];
    }

    return orig;
  }

  function result(target, path, value, merge) {
    if (merge && isPlainObject(target[path]) && isPlainObject(value)) {
      target[path] = merge({}, target[path], value);
    } else {
      target[path] = value;
    }
  }

  function split$1(path, options) {
    const id = createKey(path, options);
    if (set.memo[id]) return set.memo[id];

    const char = (options && options.separator) ? options.separator : '.';
    let keys = [];
    let res = [];

    if (options && typeof options.split === 'function') {
      keys = options.split(path);
    } else {
      keys = path.split(char);
    }

    for (let i = 0; i < keys.length; i++) {
      let prop = keys[i];
      while (prop && prop.slice(-1) === '\\' && keys[i + 1]) {
        prop = prop.slice(0, -1) + char + keys[++i];
      }
      res.push(prop);
    }
    set.memo[id] = res;
    return res;
  }

  function createKey(pattern, options) {
    let id = pattern;
    if (typeof options === 'undefined') {
      return id + '';
    }
    const keys = Object.keys(options);
    for (let i = 0; i < keys.length; i++) {
      const key = keys[i];
      id += ';' + key + '=' + String(options[key]);
    }
    return id;
  }

  function isObject(val) {
    switch (typeof val) {
      case 'null':
        return false;
      case 'object':
        return true;
      case 'function':
        return true;
      default: {
        return false;
      }
    }
  }

  set.memo = {};
  var setValue = set;

  /*!
   * valite v0.5.0
   * (c) 2018-present Vitor Luiz Cavalcanti <vitorluizc@outlook.com> (https://vitorluizc.github.io)
   * Released under the MIT License.
   */

  /**
   * A valite's validator error class.
   */
  var ValidatorError = (function (Error) {
    function ValidatorError(message) {
      Error.call(this, message);
      this.name = 'ValidatorError';
      this.message = message;
    }

    if ( Error ) ValidatorError.__proto__ = Error;
    ValidatorError.prototype = Object.create( Error && Error.prototype );
    ValidatorError.prototype.constructor = ValidatorError;

    return ValidatorError;
  }(Error));

  /**
   * Check if value is a message.
   * @param value
   */

  var isMessage = function (value) {
    if (typeof value !== 'string' && value !== true) { throw new ValidatorError('Validator should return `true` or a message instead.'); }
    if (typeof value === 'string' && !value.trim()) { throw new ValidatorError('Validator message shouldn\'t be empty.'); }
    return value !== true;
  };
  /**
   * Check if value is a message schema.
   * @param value
   */

  var isMessageSchema = function (value) { return value && typeof value === 'object'; };

  /**
   * Get an Array of properties from a given object.
   */

  var keys = Object.keys;
  /**
   * Execute concurrently validators on a value and return first error message.
   * @param value
   * @param validators
   */

  var validate = function (value, validators) {
    if ( validators === void 0 ) validators = [];

    return new Promise(function ($return, $error) {
    var execute, messages;
    execute = function (validator) { return validator(value); };
    return Promise.resolve(Promise.all(validators.map(execute))).then(function ($await_1) {
      try {
        messages = $await_1;
        return $return(messages.find(isMessage) || null);
      } catch ($boundEx) {
        return $error($boundEx);
      }
    }, $error);
  });
  };
  /**
   * Validate the whole schema concurrenly executing validators for every property
   * and returning a mirror schema with first error message for every property.
   * @param object
   * @param schema
   */

  var validateSchema = function (object, schema) { return new Promise(function ($return, $error) {
    var errors, toResolution;
    errors = Object.create(null);
    toResolution = function (property) { return new Promise(function ($return, $error) {
      var value, error;
      value = getValue(object, property);
      return Promise.resolve(validate(value, schema[property])).then(function ($await_2) {
        try {
          error = $await_2;
          setValue(errors, property, error);
          return $return();
        } catch ($boundEx) {
          return $error($boundEx);
        }
      }, $error);
    }); };
    return Promise.resolve(Promise.all(keys(schema).map(toResolution))).then(function ($await_3) {
      try {
        return $return(errors);
      } catch ($boundEx) {
        return $error($boundEx);
      }
    }, $error);
  }); };
  /**
   * Check if error message or error message schema is empty, and therefore valid.
   * @param error
   */

  var isValid$1 = function (error) {
    if (!isMessageSchema(error)) { return !isMessage(error); }

    var isError = function (property) { return typeof error[property] === 'string'; };

    return !keys(error).some(isError);
  };

  /**
   * Defines a read-only property on target object.
   * @param {Object.<*, *>} target
   * @param {string} key
   * @param {function():any} get
   */
  var defineReadOnlyProperty = function (target, key, get) { return Object.defineProperty(target, key, {
    get: get,
    set: function () {
      throw new Error('Cannot assign to read only property.');
    }
  }); };

  /**
   * A function to check if value attends to predicate.
   * @callback Predicate
   * @param {*} value Any value could be checked.
   * @returns {boolean} `true` if value attends to predicate.
   */

  /**
   * Value returned by `typeof` to determine what is value's type.
   * @typedef {('number' | 'object' | 'string' | 'symbol' | 'boolean' | 'function' | 'undefined')} Type
   */

  /**
   * Check if value is of type.
   * @param {Type} type
   * @returns {Predicate}
   */
  var isType = function (type) { return function (value) { return typeof value === type; }; };
  /**
   * Check if value is an object.
   * @param {*} value
   * @returns {boolean}
   */

  var isObject$1 = function (value) { return !!value && isType('object')(value); };
  /**
   * Check if value is not an empty string.
   * @param {*} value
   * @returns {boolean}
   */

  var isNotEmptyString = function (value) { return isType('string')(value) && !!value.trim(); };
  /**
   * Check if value is an Array and every element attends the predicate.
   * @param {Predicate} predicate
   * @returns {Predicate}
   */

  var isEvery = function (predicate) { return function (value) { return Array.isArray(value) && value.every(predicate); }; };
  /**
   * Check if every object properties attends the predicate.
   * @param {Predicate} predicate
   * @returns {Predicate}
   */

  var isEveryProperty = function (predicate) { return function (value) {
    if (!isObject$1(value)) { return false; }
    var properties = Object.keys(value).map(function (key) { return value[key]; });
    return properties.every(isEvery(predicate));
  }; };

  //
  /**
   * Model of global form accessor object.
   * @typedef {Object} FormAcessor
   * @property {Object.<string, *>} fields
   * @property {Object.<string, string>} errors
   * @property {function(string,*):void} update
   * @property {boolean} isValid
   * @property {boolean} isLoading
   * @property {function():Promise.<void>} validateForm
   * @property {function(string):Promise.<void>} validateField
   */

  var script = {
    props: {
      name: {
        type: String,
        required: true,
        validator: isNotEmptyString
      },
      model: {
        type: Object,
        required: true,
        validator: isEveryProperty(isType('function'))
      }
    },

    data: function data() {
      return {
        ticks: 0,
        errors: Object.create(null),
        values: Object.create(null)
      };
    },

    watch: {
      name: {
        handler: 'rename',
        immediate: true
      }
    },
    computed: {
      fields: function fields() {
        var this$1 = this;

        var descriptors = {};
        Object.keys(this.model).forEach(function (name) {
          descriptors[name] = {
            get: function () { return this$1.values[name]; },
            set: function (value) { return this$1.update(name, value); }
          };
        });
        return Object.create(null, descriptors);
      },

      isLoading: function isLoading() {
        return this.ticks > 0;
      },

      isValid: function isValid$1$$1() {
        return !this.isLoading && isValid$1(this.errors);
      }

    },
    methods: {
      /**
       * When name is change it swiches registered form name.
       * @param {string} to
       * @param {string} [from]
       */
      rename: function rename(to, from) {
        if (isNotEmptyString(from)) { this.$form.unregister(from); }
        this.$form.register(to, this.createFormAcessor());
      },

      /**
       * Updates a field and trigger its validation.
       * @param {string} field Field's name.
       * @param {*} value Field's value.
       */
      update: function update(field, value) {
        this.$set(this.values, field, value);
        this.validateField(field);
      },

      /**
       * Validates a field and updates Form properties.
       * @param {string} field Field's name.
       * @returns {Promise.<void>}
       */
      validateField: function validateField(field) {
        return new Promise(function ($return, $error) {
          var value, validators;
          value = this.values[field];
          validators = this.model[field];

          var $Try_1_Post = function () {
            try {
              return $return();
            } catch ($boundEx) {
              return $error($boundEx);
            }
          };

          var $Try_1_Catch = function (error) {
            try {
              this.ticks -= 1;
              console.error(("Can't validate field \"" + field + "\"."));
              return $Try_1_Post();
            } catch ($boundEx) {
              return $error($boundEx);
            }
          }.bind(this);

          try {
            var error$1;
            this.ticks += 1;
            return Promise.resolve(validate(value, validators)).then(function ($await_3) {
              try {
                error$1 = $await_3;
                this.$set(this.errors, field, error$1);
                this.ticks -= 1;
                return $Try_1_Post();
              } catch ($boundEx) {
                return $Try_1_Catch($boundEx);
              }
            }.bind(this), $Try_1_Catch);
          } catch (error) {
            $Try_1_Catch(error);
          }
        }.bind(this));
      },

      /**
       * Validates all fields and updates Form properties.
       * @returns {Promise.<void>}
       */
      validateForm: function validateForm() {
        return new Promise(function ($return, $error) {
          var $Try_2_Post = function () {
            try {
              return $return();
            } catch ($boundEx) {
              return $error($boundEx);
            }
          };

          var $Try_2_Catch = function (error) {
            try {
              this.ticks -= 1;
              console.error('Can\'t validate form.');
              return $Try_2_Post();
            } catch ($boundEx) {
              return $error($boundEx);
            }
          }.bind(this);

          try {
            var errors;
            this.ticks += 1;
            return Promise.resolve(validateSchema(this.values, this.model)).then(function ($await_4) {
              try {
                errors = $await_4;
                this.errors = errors;
                this.ticks -= 1;
                return $Try_2_Post();
              } catch ($boundEx) {
                return $Try_2_Catch($boundEx);
              }
            }.bind(this), $Try_2_Catch);
          } catch (error) {
            $Try_2_Catch(error);
          }
        }.bind(this));
      },

      /**
       * Creates the form acessor object.
       * @returns {FormAcessor}
       */
      createFormAcessor: function createFormAcessor() {
        var this$1 = this;

        var acessor = {
          update: this.update.bind(this),
          validateForm: this.validateForm.bind(this),
          validateField: this.validateField.bind(this)
        };
        defineReadOnlyProperty(acessor, 'fields', function () { return this$1.fields; });
        defineReadOnlyProperty(acessor, 'errors', function () { return this$1.errors; });
        defineReadOnlyProperty(acessor, 'isValid', function () { return this$1.isValid; });
        defineReadOnlyProperty(acessor, 'isLoading', function () { return this$1.isLoading; });
        return acessor;
      }

    },

    beforeDestroy: function beforeDestroy() {
      this.$form.unregister(this.name);
    }

  };

  /* script */
              var __vue_script__ = script;
              
  /* template */
  var __vue_render__ = function() {
    var _vm = this;
    var _h = _vm.$createElement;
    var _c = _vm._self._c || _h;
    return _c(
      "form",
      [
        _vm._t("default", null, {
          errors: _vm.errors,
          fields: _vm.fields,
          isValid: _vm.isValid,
          isLoading: _vm.isLoading,
          validateForm: _vm.validateForm,
          validateField: _vm.validateField
        })
      ],
      2
    )
  };
  var __vue_staticRenderFns__ = [];
  __vue_render__._withStripped = true;

    /* style */
    var __vue_inject_styles__ = undefined;
    /* scoped */
    var __vue_scope_id__ = undefined;
    /* module identifier */
    var __vue_module_identifier__ = undefined;
    /* functional template */
    var __vue_is_functional_template__ = false;
    /* component normalizer */
    function __vue_normalize__(
      template, style, script$$1,
      scope, functional, moduleIdentifier,
      createInjector, createInjectorSSR
    ) {
      var component = (typeof script$$1 === 'function' ? script$$1.options : script$$1) || {};

      // For security concerns, we use only base name in production mode.
      component.__file = "/Users/vitorcavalcanti/Projects/vue-form-container/src/FormContainer.vue";

      if (!component.render) {
        component.render = template.render;
        component.staticRenderFns = template.staticRenderFns;
        component._compiled = true;

        if (functional) { component.functional = true; }
      }

      component._scopeId = scope;

      

      return component
    }
    /* style inject */
    
    /* style inject SSR */
    

    
    var FormContainer = __vue_normalize__(
      { render: __vue_render__, staticRenderFns: __vue_staticRenderFns__ },
      __vue_inject_styles__,
      __vue_script__,
      __vue_scope_id__,
      __vue_is_functional_template__,
      __vue_module_identifier__,
      undefined,
      undefined
    )

  /**
   * Model of global form accessor object.
   * @typedef {Object} FormAcessor
   * @property {Object.<string, *>} fields
   * @property {Object.<string, string>} errors
   * @property {function(string,*):void} update
   * @property {boolean} isValid
   * @property {boolean} isLoading
   * @property {function():Promise.<void>} validateForm
   * @property {function(string):Promise.<void>} validateField
   */

  /**
   * Globally register FormContainer and it's helpers on Vue application.
   * @param {Vue} Vue
   */

  var install = function (Vue) {
    /**
     * State of registered forms.
     * @type {Object.<string, FormAcessor>}
     */
    var forms = Object.create(null);
    /**
     * Reaches a registered FormContainer.
     * @param {string} name
     */

    var form = function (name) {
      if (!name in forms) { throw new Error(("Can't find form \"" + name + "\".")); }
      return forms[name];
    };
    /**
     * Register a form to use it's helpers on every component.
     * @param {string} name Name used to set form and it's functions.
     * @param {FormAcessor} form
     */


    form.register = function (name, form) {
      forms[name] = form;
    };
    /**
     * Unregister form.
     * @param {string} name
     */


    form.unregister = function (name) {
      delete forms[name];
    };

    defineReadOnlyProperty(Vue.prototype, '$form', function () { return form; });
    Vue.component('FormContainer', FormContainer);
  };

  return install;

})));
